"The abstract class for the Pavlova class"

from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar, Mapping, Optional, Tuple

T = TypeVar('T')  # pylint: disable=invalid-name


class BasePavlova(ABC):
    "The base pavlova class. Use the pavlova.Pavlova class instead"
    @abstractmethod
    def from_mapping(self,
                     input_mapping: Mapping[Any, Any],
                     model_class: Type[T],
                     path: Optional[Tuple[str, ...]] = None) -> T:
        "Parse from a Mapping and return a model_class"
        pass

    @abstractmethod
    def parse_field(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> Any:
        "Parse a particular field with type field_type"
        pass



"Allows you to use Pavlova effortlessly from Flask"

from functools import wraps
from typing import Any, Callable, Dict, Type, TypeVar

import flask

from pavlova import Pavlova


T = TypeVar('T')  # pylint: disable=invalid-name


class FlaskPavlova(Pavlova):
    "The flask adaptor for Pavlova"

    def use(self, model_class: Type[T]) -> Callable:
        """Wraps a flask endpoint, parses the data coming in via json or form
        data, then passes it to the function as an argument.
        """
        def _wrapper(func: Callable) -> Callable:
            @wraps(func)
            def wrap(*args: Any, **kwargs: Dict[Any, Any]) -> Any:
                new_args = list(args)
                new_args.append(self._from_flask_request(model_class))
                return func(*new_args, **kwargs)
            return wrap
        return _wrapper

    def _from_flask_request(self, model_class: Type[T]) -> T:
        json_body: Dict[str, Any] = {}
        if flask.request.is_json:
            json_body = flask.request.get_json()

        return self.from_mapping(
            {
                **json_body,
                **flask.request.values.to_dict()
            },
            model_class
        )



"Pavlova"

#pylint: disable=no-name-in-module,ungrouped-imports

import datetime
from decimal import Decimal
from enum import Enum
import inspect
import typing
from typing import (
    Any, Dict, Type, TypeVar, Union, Generic, List, Mapping, Optional, Tuple
)
import sys

import dataclasses

from pavlova.base import BasePavlova
from pavlova.parsers import PavlovaParser
import pavlova.parsers


if sys.version_info < (3, 7):
    from typing import GenericMeta as GenericAlias  # type: ignore
else:
    from typing import _GenericAlias as GenericAlias  # type: ignore

T = TypeVar('T')  # pylint: disable=invalid-name


class PavlovaParsingError(Exception):
    """The exception that will be thrown if there is a ValueError or TypeError
    encountered when parsing a mapping."""
    def __init__(self,
                 message: str,
                 original_exception: Exception,
                 path: Tuple[str, ...],
                 expected_type: Type) -> None:
        super().__init__(message)

        self.original_exception = original_exception
        self.path = path
        self.expected_type = expected_type


class Pavlova(BasePavlova):
    "The main Pavlova class that handles parsing dictionaries"

    parsers: Dict[Any, PavlovaParser] = {}

    def __init__(self) -> None:
        self.parsers = {
            bool: pavlova.parsers.BoolParser(self),
            datetime.datetime: pavlova.parsers.DatetimeParser(self),
            float: pavlova.parsers.FloatParser(self),
            int: pavlova.parsers.IntParser(self),
            str: pavlova.parsers.StringParser(self),
            Decimal: pavlova.parsers.DecimalParser(self),
            Dict: pavlova.parsers.DictParser(self),
            Enum: pavlova.parsers.EnumParser(self),
            List: pavlova.parsers.ListParser(self),
            Union: pavlova.parsers.UnionParser(self),
        }

    def register_parser(
            self,
            parser_type: Type[Any],
            parser: pavlova.parsers.PavlovaParser,
    ) -> None:
        """Adds a PavlovaParser for a particular type, this is in addition to
        the built in parsers. If you pass in a type that is already handled by
        Pavlova, it will overwrite the built in parser.
        """
        self.parsers[parser_type] = parser

    def from_mapping(self,
                     input_mapping: Mapping[Any, Any],
                     model_class: Type[T],
                     path: Optional[Tuple[str, ...]] = None) -> T:
        """Given a dictionary and a dataclass, return an instance of the
        dataclass"""
        if path is None:
            path = tuple()

        if not dataclasses.is_dataclass(model_class):
            raise TypeError("The root class must be a dataclass")

        data = dict()
        for field in dataclasses.fields(model_class):
            if field.name not in input_mapping:
                # Check if there is a default value set. If there isn't, raise
                # an error, else continue parsing.
                if not hasattr(model_class, field.name):
                    raise PavlovaParsingError(
                        f'Field: {field.name} missing',
                        TypeError(),
                        path + (field.name,),
                        field.type,
                    )
                continue

            try:
                data[field.name] = self.parse_field(
                    input_mapping[field.name],
                    field.type,
                    path + (field.name,),
                )
            except (ValueError, TypeError) as exc:
                raise PavlovaParsingError(
                    str(exc),
                    exc,
                    path + (field.name,),
                    field.type,
                )

        return model_class(**data)  # type: ignore

    def parse_field(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> Any:
        # pylint: disable=protected-access

        if field_type in self.parsers:
            return self.parsers[field_type].parse_input(
                input_value, field_type, path
            )

        # If the type is a dataclass, go ahead and call from_mapping
        # recursively.
        if dataclasses.is_dataclass(field_type):
            return self.from_mapping(
                input_value, field_type, path
            )

        # In Python 3.7, some types, such as List, Dict, Union etc show up as
        # type '_GenericAlias'. As such, it is very hacky to track what their
        # types actually are, and what the calling party is intending.
        # In Python 3.6, the types aren't _GenericAlias, but are sometimes
        # GenericMeta, or some weird type that appears to be the same thing,
        # but isn't (Looking at you, Union)
        if field_type.__module__ == 'typing':
            if getattr(field_type, '_name', None):
                base_type = getattr(
                    sys.modules.get(field_type.__module__),
                    field_type._name,
                )
            elif hasattr(field_type, '__origin__'):
                base_type = field_type.__origin__

            return self.parsers[base_type].parse_input(
                input_value, field_type, path
            )

        # Check to see if any of the type's parent types is something we can
        # parse. This happens after the generic type checking, as those types
        # do not have a class, and therefore don't have a module resolution
        # order.
        candidate_types = [
            t for t in inspect.getmro(field_type) if t in self.parsers
        ]
        # if there is something we can use, use the most specific type, which
        # will be the first item in the list
        if candidate_types:
            return self.parsers[candidate_types[0]].parse_input(
                input_value, field_type, path
            )

        raise TypeError(f'Type {field_type} is not supported')



#pylint: disable=too-few-public-methods,missing-docstring
#pylint: disable=unused-argument,no-self-use
"This module contains all of the built in parsers for Pavlova"

from abc import ABC, abstractmethod
import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, List, Dict, Union, Type, TypeVar, Generic, Tuple

import dateparser

from pavlova.base import BasePavlova


T = TypeVar('T')  # pylint: disable=invalid-name


class PavlovaParser(Generic[T], ABC):
    "The base pavlova parser for types"

    def __init__(self, pavlova_instance: BasePavlova) -> None:
        self.pavlova = pavlova_instance

    @abstractmethod
    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> T:
        "Given an input, return it's typed value"
        pass


class BoolParser(PavlovaParser[bool]):
    "Parses a Boolean"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> bool:
        if isinstance(input_value, str):
            input_string = input_value.lower()
            if input_string in ('yes', 'true', '1'):
                return True
            if input_string in ('no', 'false', '0'):
                return False

            raise TypeError(f'{input_string} is not a valid boolean value')

        return bool(input_value)


class ListParser(PavlovaParser[List[T]]):
    "Parses a List"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> List[T]:
        if not isinstance(input_value, list):
            raise TypeError(f'Input value: {input_value} is not a list')

        sub_type = field_type.__args__[0]
        return [
            self.pavlova.parse_field(f, sub_type, path + (f'[{i}]',))
            for i, f in enumerate(input_value)
        ]


class IntParser(PavlovaParser[int]):
    "Parses ints"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> int:
        return int(input_value)


class FloatParser(PavlovaParser[float]):
    "Parses floats"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> float:
        return float(input_value)


class DecimalParser(PavlovaParser[Decimal]):
    "Parses floats"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> Decimal:
        return Decimal(input_value)


class StringParser(PavlovaParser[str]):
    "Parses a String"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> str:
        return str(input_value)


class DictParser(PavlovaParser[Dict]):
    "Parses a dictionary"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> Dict:
        'Parses a dictionary into a typed structure'
        if not isinstance(input_value, dict):
            raise TypeError(f'Input value: {input_value} is not a dict')

        key_type = field_type.__args__[0]
        value_type = field_type.__args__[1]
        return {
            self.pavlova.parse_field(k, key_type, path):
            self.pavlova.parse_field(v, value_type, path + (k,))
            for k, v in input_value.items()
        }


class DatetimeParser(PavlovaParser[datetime.datetime]):
    "Parses a datetime"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> datetime.datetime:
        return dateparser.parse(input_value)


class UnionParser(PavlovaParser[Union[T]]):
    "Parses an Union"

    @staticmethod
    def _is_from_optional(field_type: Type) -> bool:
        if not hasattr(field_type, '__args__'):
            return False

        if len(field_type.__args__) != 2:
            return False

        none_type: Any = type(None)
        if not field_type.__args__[1] is none_type:
            return False

        return True

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> T:
        if not self._is_from_optional(field_type):
            raise TypeError('Unions of this type are not allowed')

        real_field_type = field_type.__args__[0]
        return self.pavlova.parse_field(input_value, real_field_type, path)


class EnumParser(PavlovaParser[Enum]):
    "Parses enums"

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> Enum:
        # If the input is a string, try matching it against the names of the
        # enum members. For consistency, we will lower case both values first.
        if isinstance(input_value, str):
            enum_name = input_value.lower()
            enum_values = [
                m for m in field_type if m.name.lower() == enum_name
            ]
            if enum_values:
                return enum_values[0]

        # Try instantiating the enum with our value
        return field_type(input_value)


class GenericParser(PavlovaParser[T]):
    def __init__(self, pavlova: BasePavlova, parser_type: T) -> None:
        super().__init__(pavlova)
        self.parser_type = parser_type

    def parse_input(self,
                    input_value: Any,
                    field_type: Type,
                    path: Tuple[str, ...]) -> T:
        return self.parser_type(input_value)  # type: ignore



from setuptools import find_packages, setup
import sys

requirements = [
    'dateparser',
]

if sys.version_info < (3, 7):
    requirements.append('dataclasses')

setup(
    name='pavlova',
    version='0.1.2',
    description='Simplified deserialization using dataclasses',
    long_description=open('README.rst').read(),
    author='Freelancer.com',
    author_email='chris@freelancer.com',
    url='https://github.com/freelancer/pavlova',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    install_requires=requirements,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser '
        'General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.7',
    ]
)
